"""
Use the GitHub API to show the status of current GitHub Action runs on a
specific branch.

It tries to show enough runs to see the latest status of each workflow.  Runs
older than a week are not considered.

"""

import contextlib
import io
import os.path
import re
import signal
import sys
import urllib.parse

import click
import exceptiongroup
import rich.console

from .data_core import draw_runs
from .git_help import git_repo_urls, git_branch
from .http_help import get_data
from .utils import Interval, WatchGhaError


console = rich.console.Console(highlight=False)
error_console = rich.console.Console(stderr=True, highlight=False)


def fatal(msg, status=2):
    error_console.print(msg)
    sys.exit(status)


@contextlib.contextmanager
def handle_resize(handler):
    original_sigwinch_handler = None
    try:
        # Save the original handler.
        original_sigwinch_handler = signal.getsignal(signal.SIGWINCH)
        # Call the given handler on sigwinch.
        signal.signal(signal.SIGWINCH, lambda _, __: handler())
    except AttributeError:
        # This system is probably windows, and doesn't have SIGWINCH.
        # Unfortunately there's no signal for window resize on windows and I
        # don't know how to properly handle it. Just swallow the error.
        pass

    # yield control to the with block, and re-raise any exceptions
    try:
        yield
    except:
        raise
    finally:
        # clean up our sigwinch handler if we set one
        if original_sigwinch_handler:
            signal.signal(signal.SIGWINCH, original_sigwinch_handler)


@click.command()
@click.option("--sha", help="The commit SHA to use. Must be a full SHA.")
@click.option(
    "--poll",
    help="How many seconds between refreshes.",
    type=int,
    default=15,
    show_default=True,
)
@click.option("--wait-for-start", is_flag=True, help="Wait for jobs to start.")
@click.option(
    "--only",
    help=(
        "Words to limit the workflows shown. "
        + "Only workflows with these comma separated case insensitive substrings "
        + "in their names will be shown."
    ),
)
@click.argument("repo", default=".")
@click.argument("branch", required=False)
def main(sha, poll, wait_for_start, only, repo, branch):
    """
    Watch GitHub Action runs.

    Repeatedly gets the latest status and redraws the screen, until all of the
    jobs are complete.

    REPO is a local directory or GitHub URL, defaulting to ".".

    BRANCH is defaulted from the git repo.

    """
    if only is not None:
        only_words = [w.strip() for w in only.split(",")]
    else:
        only_words = None

    watcher = GhaWatcher(
        urls=gha_urls(repo, branch, sha),
        get_data_fn=get_data,
        only_words=only_words,
    )

    watcher.watch(wait_for_start, poll, console)


def gha_urls(repo, branch, sha):
    """Figure out the GHA api URLs to use for `repo`, `branch`, and `sha`."""
    if os.path.isdir(repo):
        repo_urls = list(git_repo_urls(repo))
        if branch is None:
            branch = git_branch(repo)
    elif ":" in repo:
        repo_urls = [repo]
        if branch is None:
            fatal(f"Branch is required for URL repo")
    else:
        fatal(f"Don't understand repo {repo!r}")

    params = {"per_page": "40"}
    if sha:
        params["head_sha"] = sha
    else:
        assert branch is not None
        params["branch"] = branch
    url_args = urllib.parse.urlencode(params)

    github_urls = []
    for repo_url in repo_urls:
        # repo_url = "https://github.com/owner/repo.git"
        # repo_url = "git@github.com:someorg/somerepo.git"
        # see also https://docs.github.com/en/actions/learn-github-actions/variables#default-environment-variables
        server_url = os.getenv("GITHUB_SERVER_URL", "https://github.com")
        repo_match = re.fullmatch(
            rf"(?:{re.escape(server_url)}/|git@github.com:)([^/]+/[^/]+?)(?:\.git|/)?",
            repo_url,
        )
        if repo_match is None:
            continue

        api_url = os.getenv("GITHUB_API_URL", "https://api.github.com")
        url = f"{api_url}/repos/{repo_match[1]}/actions/runs?{url_args}"
        github_urls.append(url)

    if not github_urls:
        fatal(f"Couldn't find GitHub repo from remote urls: {repo_urls!r}")

    return github_urls


class GhaWatcher:
    def __init__(self, urls, get_data_fn, only_words):
        self.urls = urls
        self.get_data_fn = get_data_fn
        self.only_words = only_words
        self.status = 0
        self.error = None

    def watch(self, wait_for_start, poll, console):
        self.done = False
        self.succeeded = False
        self.interrupted = False

        self.watch_gha_errors = []
        interval = Interval(poll)

        output = ""

        with exceptiongroup.catch(
            {
                WatchGhaError: self.handle_watchghaerror,
                KeyboardInterrupt: self.handle_keyboardinterrupt,
            }
        ):
            while True:
                output = self.get_gha_display()
                if wait_for_start:
                    if not self.done:
                        break
                else:
                    break

            if not self.done:
                with console.screen() as screen:
                    with handle_resize(lambda: screen.update(output)):
                        while not self.done:
                            screen.update(output)
                            interval.wait()
                            output = self.get_gha_display()

        if self.watch_gha_errors:
            fatal(self.watch_gha_errors[0])
        console.print(output, end="")
        if self.interrupted:
            fatal("** interrupted **", status=2)
        sys.exit(0 if self.succeeded else 1)

    def handle_watchghaerror(self, excgroup):
        self.watch_gha_errors.extend(excgroup.exceptions)

    def handle_keyboardinterrupt(self, excgroup):
        self.interrupted = True

    def get_gha_display(self):
        stream = io.StringIO()

        self.done, self.succeeded = draw_runs(
            self.urls,
            only_words=self.only_words,
            datafn=self.get_data_fn,
            outfn=lambda s: print(s, file=stream),
        )
        return stream.getvalue()



"""Data collection and rendering for watchgha."""

import datetime
import itertools
import json

import trio

from .bucketer import DatetimeBucketer
from .utils import nice_time, to_datetime, DictAttr


bucketer = DatetimeBucketer(5)

CSTYLES = {
    "failure": "red bold",
    "startup_failure": "red bold",
    "success": "green bold",
}

# States that are finished.
FINISHED = {
    "cancelled",
    "failure",
    "skipped",
    "startup_failure",
    "success",
}

# States that don't need jobs displayed.
NO_JOBS = {
    "skipped",
    "success",
}

CICONS = {
    "cancelled": "\N{DAGGER}",
    "failure": "\N{BALLOT X}",
    "in_progress": "\N{CLOCKWISE OPEN CIRCLE ARROW}",
    "pending": "\N{TIMER CLOCK}",
    "queued": "\N{TIMER CLOCK}",
    "skipped": "\N{BALLOT BOX}",
    "startup_failure": "\N{BALLOT X}",
    "success": "\N{CHECK MARK}",
}

STEPDOTS = {
    "cancelled": "[red]\N{DAGGER}[/]",
    "failure": "[red]\N{BULLET}[/]",
    "in_progress": "[white]\N{BULLET}[/]",
    "pending": "[dim white]\N{BULLET}[/]",
    "queued": "[dim white]\N{BULLET}[/]",
    "skipped": "[default]\N{WHITE BULLET}[/]",
    "success": "[green]\N{BULLET}[/]",
}


def summary_style_icon(data):
    summary = data["status"]
    if summary == "completed":
        summary = data["conclusion"]
    style = CSTYLES.get(summary, "default")
    icon = CICONS.get(summary, " ")
    return summary, style, icon


def run_group_key(run_data):
    return (
        bucketer.defuzz(run_data["started_dt"]),
        run_data["head_sha"],
        run_data["event"],
    )


def run_sort_key(run_data):
    return (
        bucketer.defuzz(run_data["started_dt"]),
        run_data["head_sha"],
        run_data["event"],
        run_data["name"],
    )


def draw_runs(urls, only_words, datafn, outfn):
    # Workflow runs is a flat list of runs.  We bucket them by time started,
    # sha, and event to create "events".  Each event has a number of runs, each
    # run has a number of jobs, each job has a number of steps. They end up
    # displayed as:
    #
    #   event-name sha, time
    #       outcome run-name, url
    #           job-name     current-step-or-outcome

    events = trio.run(get_events, urls, only_words, datafn)
    done, succeeded = draw_events(events, outfn)
    return done, succeeded


async def get_events(urls, only_words, datafn):
    runs = []

    async def runs_from_url(url):
        runs.extend(json.loads(await datafn(url))["workflow_runs"])

    async with trio.open_nursery() as nursery:
        for url in urls:
            nursery.start_soon(runs_from_url, url)

    # In odd situations (duplicate remotes) we can get the same run more than
    # once. De-duplicate them.
    runs_by_id = {r["id"]: r for r in runs}
    runs = list(runs_by_id.values())

    async with trio.open_nursery() as nursery:
        for run in runs:
            run["started_dt"] = to_datetime(run["run_started_at"])

        runs.sort(key=run_sort_key, reverse=True)
        run_names_seen = {"Cancel"}

        events = []

        for _, g in itertools.groupby(runs, key=run_group_key):
            event_runs = list(g)
            these_runs_names = {run["name"] for run in event_runs}
            # If the .yml file couldn't even be parsed, the run name is the
            # name of the .yml file.  Exclude those, or a bad parse will
            # pollute the run list.
            these_runs_names = {
                n for n in these_runs_names if not n.startswith(".github/")
            }
            if not (these_runs_names - run_names_seen):
                continue
            days_old = (
                datetime.datetime.now(datetime.timezone.utc)
                - event_runs[0]["started_dt"]
            ).days
            if days_old > 7:
                continue

            if only_words is not None:
                event_runs = [
                    run
                    for run in event_runs
                    if any(word in run["name"].lower() for word in only_words)
                ]
                if not event_runs:
                    continue

            events.append(event_runs)
            run_names_seen.update(these_runs_names)

            async def load_run(run):
                run["jobs"] = json.loads(await datafn(run["jobs_url"]))["jobs"]

            for run in event_runs:
                summary, _, _ = summary_style_icon(run)
                if summary != "success":
                    nursery.start_soon(load_run, run)

    return events


def draw_events(events, outfn):
    done = True
    succeeded = True
    for event_runs in events:
        e = DictAttr(event_runs[0])
        outfn(
            f"[white bold]{e.display_title}[/] "
            + f"{e.head_branch} "
            + f"\\[{e.event}] "
            + f"  [dim]{e.head_sha:.12}  @{nice_time(e.started_dt)}[/]"
        )
        for run in event_runs:
            summary, style, icon = summary_style_icon(run)
            if summary not in FINISHED:
                done = False
            r = DictAttr(run)
            run_id = r.html_url.split("/")[-1]
            outfn(
                "   "
                + f"[{style}]{icon} {summary:12}[/] "
                + f"[white bold]{r.name:16}[/] "
                + f"  [blue link={r.html_url}]view {run_id}[/]"
            )

            if summary in NO_JOBS:
                continue

            succeeded = False
            for job in run["jobs"]:
                current_step, style, icon = summary_style_icon(job)
                stepdots = ""
                if current_step != "success":
                    if job["status"] == "queued":
                        current_step = "queued"
                        done = False
                    else:
                        steps = job["steps"]
                        for step in steps:
                            if (
                                step["status"] == "completed"
                                and step["conclusion"] == "failure"
                            ):
                                current_step = f"failure {step['name']}"
                                break
                            if step["status"] == "in_progress":
                                done = False
                                stepdots = ""
                                for s in steps:
                                    ssum = summary_style_icon(s)[0]
                                    stepdots += STEPDOTS.get(ssum, "?")
                                current_step = f" {step['name']}"
                                break
                        else:
                            if steps:
                                current_step = steps[-1]["name"]
                            elif (
                                job["status"] == "completed"
                                and job["conclusion"] == "skipped"
                            ):
                                current_step = "skipped"
                            else:
                                current_step = "-None-"

                j = DictAttr(job)
                outfn(
                    "      "
                    + f"{j.name:30} [{style}]{icon}[/] "
                    + f"{stepdots}[{style}]{current_step}[/]"
                )

    return done, succeeded



"""A module to show styling.

Run from the command line with:

    $ python3 -m watchgha.demo

"""

import json

import rich.console

from .data_core import FINISHED, draw_runs


RUN_COMMON = {
    "display_title": "fix: most awesome fix",
    "head_branch": "nedbat/test",
    "html_url": "https://github.com/owner/repo/actions/runs/123456789",
    "event": "push",
    "head_sha": "4b2ff58124791953563fdb52e40d9ab79d274d9a",
    "run_started_at": "2023-07-02T13:49:27Z",
}

DEMO_DATA = {
    "demo:one": {
        "workflow_runs": [
            {
                **RUN_COMMON,
                "name": "Test suite",
                "status": "in_progress",
                "jobs_url": "demo:jobs_tests",
            },
            *[
                {
                    **RUN_COMMON,
                    "name": f"A {conc} run",
                    "status": "completed",
                    "conclusion": conc,
                    "jobs_url": "demo:jobs_1",
                }
                for conc in FINISHED
            ],
        ],
    },
    "demo:jobs_1": {
        "jobs": [
            {
                "name": "Just one job",
                "status": "queued",
            },
        ],
    },
    "demo:jobs_tests": {
        "jobs": [
            {
                "name": "A delayed job",
                "status": "queued",
            },
            {
                "name": "A finished job",
                "status": "completed",
                "conclusion": "success",
            },
            {
                "name": "A failed job",
                "status": "completed",
                "conclusion": "failure",
                "steps": [
                    {
                        "status": "completed",
                        "conclusion": "success",
                    },
                    {
                        "status": "completed",
                        "conclusion": "failure",
                        "name": "Didn't work",
                    },
                    {
                        "status": "queued",
                    },
                ],
            },
            *[
                {
                    "name": f"Test suite Py 3.{py}",
                    "status": "in_progress",
                    "steps": [
                        *[
                            {
                                "status": "completed",
                                "conclusion": "success",
                            }
                            for _ in range(3 - py % 2)
                        ],
                        {
                            "status": "skipped",
                        },
                        {
                            "name": "Prep tests" if (py % 2) else "Run the tests",
                            "status": "in_progress",
                        },
                        *[
                            {
                                "status": "queued",
                            }
                            for _ in range(2 + py % 2)
                        ],
                    ],
                }
                for py in [8, 9, 10, 11]
            ],
        ],
    },
}


async def demo_datafn(url):
    return json.dumps(DEMO_DATA[url])


def demo():
    console = rich.console.Console(highlight=False)
    draw_runs("demo:one", datafn=demo_datafn, outfn=console.print)


if __name__ == "__main__":
    demo()



"""
Get information from local git.
"""

import functools

import dulwich.porcelain
import dulwich.repo


@functools.lru_cache(maxsize=10)
def _dulwich_repo(dir="."):
    return dulwich.repo.Repo(dir)


def git_repo_urls(dir):
    """Find all the remote URLs for a git repo at `dir`."""
    config = _dulwich_repo(dir).get_config()
    for section in config.sections():
        if section[0] == b"remote":
            url = config.get(section, "url").decode()
            yield url


def git_branch(dir):
    """Get the current git branch name."""
    return dulwich.porcelain.active_branch(_dulwich_repo(dir)).decode()



import datetime
import time


class WatchGhaError(Exception):
    pass


def nice_time(dt):
    dt = dt.astimezone()
    now = datetime.datetime.now()
    if dt.date() != now.date():
        if dt.year != now.year:
            fmt = "%Y-%m-%d %I:%M%p"
        else:
            fmt = "%m-%d %I:%M%p"
    else:
        fmt = "%I:%M%p"
    return dt.strftime(fmt).lower()


def to_datetime(isostr):
    # 3.11 accepts Z, but older Pythons don't.
    isostr = isostr.replace("Z", "+00:00")
    return datetime.datetime.fromisoformat(isostr)


class DictAttr:
    def __init__(self, d):
        self.d = d

    def __getattr__(self, name):
        return self.d[name]


class Interval:
    """Wait for an interval of time to pass.

    Better than time.sleep because it accounts for time spent doing things
    other than sleeping.

    """

    def __init__(self, secs):
        self.secs = secs
        self.last_time = time.time()

    def wait(self):
        now = time.time()
        delay = self.secs - (now - self.last_time)
        self.last_time = now
        if delay > 0:
            time.sleep(delay)



"""
Helper for getting data from URLs.
"""

import itertools
import mimetypes
import os

import httpx
import trio

from .utils import WatchGhaError


RETRY_STATUS_CODES = {502}


class Http:
    """
    Helper for getting data from URLs.

    Uses the GITHUB_TOKEN environment variable (if set) as authentication.

    Define SAVE_DATA=1 in the environment to save retrieved data in get_*.*
    files.

    """

    def __init__(self):
        # $set_env.py: SAVE_DATA - save all fetched data to get_* files.
        self.save = bool(int(os.environ.get("SAVE_DATA", "0")))
        if self.save:
            with open("get_index.txt", "w") as index:
                index.write("# URLs fetched:\n")
            self.count = itertools.count()
        self.auth = None
        self.headers = {}
        token = os.environ.get("GITHUB_TOKEN", "")
        if token:
            self.headers["Authorization"] = f"Bearer {token}"
        else:
            try:
                self.auth = httpx.NetRCAuth()
            except FileNotFoundError:
                self.auth = None

    async def get_data(self, url):
        async with httpx.AsyncClient(auth=self.auth) as client:
            resp = None
            try:
                for ntry in range(3):
                    resp = await client.get(
                        url,
                        headers=self.headers,
                        timeout=30,
                        follow_redirects=True,
                    )
                    if resp.status_code not in RETRY_STATUS_CODES:
                        break
                    await trio.sleep(0.05 * 2**ntry)
                resp.raise_for_status()
            except httpx.HTTPError as e:
                # Some error messages have the URL, and some don't.  Add it in
                # if it isn't there already.
                if len(url) > 10 and url in str(e):
                    msg = str(e)
                else:
                    msg = f"Couldn't get {url!r}: {e}"
                if resp is not None:
                    try:
                        for label, text in resp.json().items():
                            msg += f"\n{label}: {text}"
                    except Exception:
                        msg += f"\n{resp.text}"
                raise WatchGhaError(msg) from e
            data = resp.text
            if self.save:
                ext = extension_for_content(resp)
                filename = f"get_{next(self.count):03d}{ext}"
                async with await trio.open_file("get_index.txt", "a") as index:
                    await index.write(f"{filename}: {url}\n")
                async with await trio.open_file(filename, "w") as out:
                    await out.write(data)
            return data


def extension_for_content(response):
    content_type = response.headers["content-type"].partition(";")[0].strip()
    return mimetypes.guess_extension(content_type)


_get_data = Http().get_data


async def get_data(*args, **kwargs):
    """Run get_data three times, with retry."""
    # I don't like that this has a retry loop and get_data above also does.
    for ntry in range(3):
        try:
            return await _get_data(*args, **kwargs)
        except Exception as exc:
            exc_to_raise = exc
            await trio.sleep(0.05 * 2**ntry)
    raise exc_to_raise



class DatetimeBucketer:
    def __init__(self, window):
        self.window = window
        # The set of good instants
        self.instants = set()
        # Map rounded times to good instants
        self.rounds = {}

    def roundings(self, dt):
        for jitter in [0, self.window / 2]:
            yield round((dt.timestamp() + jitter) / self.window)

    def defuzz(self, dt):
        if dt in self.instants:
            return dt

        for rounded in self.roundings(dt):
            instant = self.rounds.get(rounded)
            if instant is not None:
                return instant

        self.instants.add(dt)
        for rounded in self.roundings(dt):
            self.rounds[rounded] = dt

        return dt
